//! Code which enables sending messages between processes in the blocktree system.
use btlib::{crypto::rand_array, error::BoxInIoErr, BlockPath, Result, Writecap};
use btserde::{read_from, write_to};
use bytes::{BufMut, BytesMut};
use core::{
    future::Future,
    pin::Pin,
    task::{Context, Poll},
};
use futures::{
    sink::{Send, Sink},
    stream::Stream,
    SinkExt, StreamExt,
};
use lazy_static::lazy_static;
use serde::{Deserialize, Serialize};
use std::{io, net::Shutdown, path::PathBuf};
use tokio::{
    io::{AsyncRead, AsyncWrite, ReadBuf},
    net::UnixDatagram,
};
use tokio_util::codec::{Decoder, Encoder, FramedRead, FramedWrite};
use zerocopy::FromBytes;

pub use private::*;

mod private {

    use super::*;

    lazy_static! {
        /// The default directory in which to place blocktree sockets.
        static ref SOCK_DIR: PathBuf = {
            let mut path: PathBuf = std::env::var("HOME").unwrap().into();
            path.push(".btmsg");
            path
        };
    }

    /// Appends the given Blocktree path to the path of the given directory.
    fn socket_path(fs_path: &mut PathBuf, block_path: &BlockPath) {
        fs_path.push(block_path.to_string());
    }

    #[derive(PartialEq, Eq, Serialize, Deserialize)]
    pub enum MsgError {
        Unknown,
    }

    /// The owned version of a message body. This is the body type for received messages. It's
    /// very important that this enum's variants match those of [MsgBodyRef], otherwise runtime
    /// deserialization errors will occur.
    #[derive(Deserialize)]
    pub enum MsgBodyOwned {
        Success,
        Fail(MsgError),
        Ping,
        Hello(Writecap),
        Read { offset: u64, size: u64 },
        Write { offset: u64, buf: Vec<u8> },
        Custom(Vec<u8>),
    }

    /// The reference version of a message body. This is the body type when sending messages. It's
    /// very important that this enum's variants match those of [MsgBodyOwned], otherwise runtime
    /// deserialization errors will occur.
    #[derive(Serialize)]
    pub enum MsgBodyRef<'a> {
        Success,
        Fail(&'a MsgError),
        Ping,
        Hello(&'a Writecap),
        Read { offset: u64, size: u64 },
        Write { offset: u64, buf: &'a [u8] },
        Custom(&'a [u8]),
    }

    /// Trait which unifies owned and borrowed messages.
    pub trait Msg {
        type Body;
        fn to(&self) -> &BlockPath;
        fn from(&self) -> &BlockPath;
        fn id(&self) -> u128;
        fn body(&self) -> &Self::Body;
    }

    /// An owned message. This is the type which observed by the receiver.
    #[derive(Deserialize)]
    pub struct MsgOwned {
        to: BlockPath,
        from: BlockPath,
        id: u128,
        body: MsgBodyOwned,
    }

    impl Msg for MsgOwned {
        type Body = MsgBodyOwned;

        fn to(&self) -> &BlockPath {
            &self.to
        }

        fn from(&self) -> &BlockPath {
            &self.from
        }

        fn id(&self) -> u128 {
            self.id
        }

        fn body(&self) -> &Self::Body {
            &self.body
        }
    }

    /// A borrowed message. This is the type which is produced by the sender.
    #[derive(Serialize)]
    pub struct MsgRef<'a> {
        to: &'a BlockPath,
        from: &'a BlockPath,
        id: u128,
        body: MsgBodyRef<'a>,
    }

    impl<'a> MsgRef<'a> {
        pub fn new(to: &'a BlockPath, from: &'a BlockPath, id: u128, body: MsgBodyRef<'a>) -> Self {
            Self { to, from, id, body }
        }
    }

    impl<'a> Msg for MsgRef<'a> {
        type Body = MsgBodyRef<'a>;

        fn to(&self) -> &BlockPath {
            self.to
        }

        fn from(&self) -> &BlockPath {
            self.from
        }

        fn id(&self) -> u128 {
            self.id
        }

        fn body(&self) -> &Self::Body {
            &self.body
        }
    }

    /// An owned message tagged with a version.
    #[derive(Deserialize)]
    enum VerMsgOwned {
        V0(MsgOwned),
    }

    /// A borrowed message tagged with a version.
    #[derive(Serialize)]
    enum VerMsgRef<'a> {
        V0(MsgRef<'a>),
    }

    /// A type which can be used to send messages.
    pub trait Sender<'a>: Sink<MsgRef<'a>, Error = btlib::Error> {
        type SendFut: Future<Output = Result<()>> + std::marker::Send;

        fn path(&self) -> &BlockPath;

        /// Creates a new message with the given `from` and `body` fields and sends it to the peer
        /// this [Sender] is associated with.
        fn send_msg(&'a mut self, from: &'a BlockPath, body: MsgBodyRef<'a>) -> Self::SendFut;

        /// Generates and returns a new message ID.
        fn gen_id() -> Result<u128> {
            const LEN: usize = std::mem::size_of::<u128>();
            let bytes = rand_array::<LEN>()?;
            let option = u128::read_from(bytes.as_slice());
            // Safety: because LEN == size_of::<u128>(), read_from should have returned Some.
            Ok(option.unwrap())
        }
    }

    /// A type which can be used to receive messages.
    pub trait Receiver: Stream<Item = Result<MsgOwned>> {
        fn path(&self) -> &BlockPath;
    }

    /// Encodes and decodes messages using [btserde].
    struct MessageCodec;

    impl<'a> Encoder<MsgRef<'a>> for MessageCodec {
        type Error = btlib::Error;

        fn encode(&mut self, item: MsgRef<'a>, dst: &mut BytesMut) -> Result<()> {
            const U64_LEN: usize = std::mem::size_of::<u64>();
            let payload = dst.split_off(U64_LEN);
            let mut writer = payload.writer();
            write_to(&VerMsgRef::V0(item), &mut writer)?;
            let payload = writer.into_inner();
            let payload_len = payload.len() as u64;
            let mut writer = dst.writer();
            write_to(&payload_len, &mut writer)?;
            let dst = writer.into_inner();
            dst.unsplit(payload);
            Ok(())
        }
    }

    impl Decoder for MessageCodec {
        type Item = MsgOwned;
        type Error = btlib::Error;

        fn decode(&mut self, src: &mut BytesMut) -> Result<Option<Self::Item>> {
            let mut slice: &[u8] = src.as_ref();
            let payload_len: u64 = match read_from(&mut slice) {
                Ok(payload_len) => payload_len,
                Err(err) => {
                    if let btserde::Error::Eof = err {
                        return Ok(None);
                    }
                    return Err(err.into());
                }
            };
            let payload_len: usize = payload_len.try_into().box_err()?;
            if slice.len() < payload_len {
                src.reserve(payload_len - slice.len());
                return Ok(None);
            }
            let VerMsgOwned::V0(msg) = read_from(&mut slice)?;
            let _ = src.split_to(std::mem::size_of::<u64>() + payload_len);
            Ok(Some(msg))
        }
    }

    /// Wraps a [UnixDatagram] and implements [AsyncRead] and [AsyncWrite] for it. Read operations
    /// are translated  to calls to `recv_from` and write operations are translated to `send`. Note
    /// that this means that writes will fail unless the wrapped socket is connected  to a peer.
    struct DatagramAdapter {
        socket: UnixDatagram,
    }

    impl DatagramAdapter {
        fn new(socket: UnixDatagram) -> Self {
            Self { socket }
        }

        fn get_ref(&self) -> &UnixDatagram {
            &self.socket
        }

        fn get_mut(&mut self) -> &mut UnixDatagram {
            &mut self.socket
        }
    }

    impl AsRef<UnixDatagram> for DatagramAdapter {
        fn as_ref(&self) -> &UnixDatagram {
            self.get_ref()
        }
    }

    impl AsMut<UnixDatagram> for DatagramAdapter {
        fn as_mut(&mut self) -> &mut UnixDatagram {
            self.get_mut()
        }
    }

    impl AsyncRead for DatagramAdapter {
        fn poll_read(
            self: Pin<&mut Self>,
            cx: &mut Context<'_>,
            buf: &mut ReadBuf<'_>,
        ) -> Poll<io::Result<()>> {
            self.socket.poll_recv(cx, buf)
        }
    }

    impl AsyncWrite for DatagramAdapter {
        fn poll_write(
            self: Pin<&mut Self>,
            cx: &mut Context<'_>,
            buf: &[u8],
        ) -> Poll<io::Result<usize>> {
            self.socket.poll_send(cx, buf)
        }

        fn poll_shutdown(self: Pin<&mut Self>, _cx: &mut Context<'_>) -> Poll<io::Result<()>> {
            Poll::Ready(self.socket.shutdown(Shutdown::Write))
        }

        fn poll_flush(self: Pin<&mut Self>, _cx: &mut Context<'_>) -> Poll<io::Result<()>> {
            Poll::Ready(Ok(()))
        }
    }

    /// Returns a [Receiver] which can be used to receive messages addressed to the given path.
    /// The `fs_path` argument specifies the filesystem directory under which the receiver's socket
    /// will be stored.
    pub fn local_receiver(fs_path: PathBuf, block_path: BlockPath) -> Result<impl Receiver> {
        UnixReceiver::new(fs_path, block_path)
    }

    /// An implementation of [Receiver] which uses a Unix datagram socket for receiving messages.
    struct UnixReceiver {
        path: BlockPath,
        socket: FramedRead<DatagramAdapter, MessageCodec>,
    }

    impl UnixReceiver {
        fn new(mut fs_path: PathBuf, block_path: BlockPath) -> Result<Self> {
            socket_path(&mut fs_path, &block_path);
            std::fs::create_dir_all(fs_path.parent().unwrap())?;
            let socket = DatagramAdapter::new(UnixDatagram::bind(fs_path)?);
            let socket = FramedRead::new(socket, MessageCodec);
            Ok(Self {
                path: block_path,
                socket,
            })
        }
    }

    impl Stream for UnixReceiver {
        type Item = Result<MsgOwned>;

        fn poll_next(mut self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Option<Self::Item>> {
            self.socket.poll_next_unpin(cx)
        }
    }

    impl Receiver for UnixReceiver {
        fn path(&self) -> &BlockPath {
            &self.path
        }
    }

    /// Returns a [Sender] which can be used to send messages to the given Blocktree path.
    /// The `fs_path` argument specifies the filesystem directory in which to locate the
    /// socket of the recipient.
    pub fn local_sender(
        fs_path: PathBuf,
        block_path: BlockPath,
    ) -> Result<impl for<'a> Sender<'a>> {
        UnixSender::new(fs_path, block_path)
    }

    /// An implementation of [Sender] which uses a Unix datagram socket to send messages.
    struct UnixSender {
        path: BlockPath,
        socket: FramedWrite<DatagramAdapter, MessageCodec>,
    }

    impl UnixSender {
        fn new(mut fs_path: PathBuf, block_path: BlockPath) -> Result<Self> {
            let socket = UnixDatagram::unbound()?;
            socket_path(&mut fs_path, &block_path);
            socket.connect(fs_path)?;
            let socket = FramedWrite::new(DatagramAdapter::new(socket), MessageCodec);
            Ok(Self {
                path: block_path,
                socket,
            })
        }
    }

    impl Sink<MsgRef<'_>> for UnixSender {
        type Error = btlib::Error;

        fn poll_ready(mut self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Result<()>> {
            self.socket.poll_ready_unpin(cx)
        }

        fn start_send(mut self: Pin<&mut Self>, item: MsgRef<'_>) -> Result<()> {
            self.socket.start_send_unpin(item)
        }

        fn poll_flush(mut self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Result<()>> {
            self.socket.poll_flush_unpin(cx)
        }

        fn poll_close(mut self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Result<()>> {
            self.socket.poll_close_unpin(cx)
        }
    }

    impl<'a> Sender<'a> for UnixSender {
        type SendFut = Send<'a, FramedWrite<DatagramAdapter, MessageCodec>, MsgRef<'a>>;

        fn path(&self) -> &BlockPath {
            &self.path
        }

        fn send_msg(&'a mut self, from: &'a BlockPath, body: MsgBodyRef<'a>) -> Self::SendFut {
            let id = Self::gen_id().unwrap();
            let msg = MsgRef::new(&self.path, from, id, body);
            self.socket.send(msg)
        }
    }

    /// Returns a connected [Sender] and [Receiver].
    pub fn local_pair(
        dir: PathBuf,
        block_path: BlockPath,
    ) -> Result<(impl for<'a> Sender<'a>, impl Receiver)> {
        let receiver = local_receiver(dir.clone(), block_path.clone())?;
        let sender = local_sender(dir, block_path)?;
        Ok((sender, receiver))
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    use tempdir::TempDir;

    lazy_static! {
        static ref BT_ROOT: BlockPath =
            BlockPath::try_from("0!dSip4J0kurN5VhVo_aTipM-ywOOWrqJuRRVQ7aa-bew").unwrap();
    }

    fn block_path<'a, I: Iterator<Item = &'a str>>(components: I) -> BlockPath {
        let mut path = BT_ROOT.clone();
        for component in components {
            path.push_component(component.to_string());
        }
        path
    }

    struct TestCase {
        dir: TempDir,
    }

    impl TestCase {
        fn new() -> TestCase {
            Self {
                dir: TempDir::new("btmsg").unwrap(),
            }
        }

        fn endpoint(&self, name: &str) -> (BlockPath, impl for<'a> Sender<'a>, impl Receiver) {
            let block_path = block_path(["apps", name].into_iter());
            let (sender, receiver) =
                local_pair(self.dir.path().to_owned(), block_path.clone()).unwrap();
            (block_path, sender, receiver)
        }
    }

    #[tokio::test]
    async fn message_received_is_message_sent() {
        let case = TestCase::new();
        let (block_path, mut sender, mut receiver) = case.endpoint("social");

        sender
            .send_msg(&block_path, MsgBodyRef::Ping)
            .await
            .unwrap();
        let actual = receiver.next().await.unwrap().unwrap();

        let matched = if let MsgBodyOwned::Ping = actual.body() {
            true
        } else {
            false
        };
        assert!(matched);
        assert_eq!(&block_path, actual.to());
        assert_eq!(&block_path, actual.from());
    }

    #[tokio::test]
    async fn ping_pong() {
        let case = TestCase::new();
        let (block_path_one, mut sender_one, mut receiver_one) = case.endpoint("one");
        let (block_path_two, mut sender_two, mut receiver_two) = case.endpoint("two");

        let handle = tokio::spawn(async move {
            let msg = receiver_one.next().await.unwrap().unwrap();
            let reply_body = if let MsgBodyOwned::Ping = msg.body() {
                MsgBodyRef::Success
            } else {
                MsgBodyRef::Fail(&MsgError::Unknown)
            };
            sender_two
                .send_msg(&block_path_one, reply_body)
                .await
                .unwrap();
        });

        sender_one
            .send_msg(&block_path_two, MsgBodyRef::Ping)
            .await
            .unwrap();
        handle.await.unwrap();
        let reply = receiver_two.next().await.unwrap().unwrap();
        let matched = if let MsgBodyOwned::Success = reply.body() {
            true
        } else {
            false
        };
        assert!(matched)
    }

    #[tokio::test]
    async fn read_write() {
        let case = TestCase::new();
        let (block_path_one, mut sender_one, mut receiver_one) = case.endpoint("one");
        let (block_path_two, mut sender_two, mut receiver_two) = case.endpoint("two");

        let handle = tokio::spawn(async move {
            let data: [u8; 8] = [0, 1, 2, 3, 4, 5, 6, 7];
            let msg = receiver_one.next().await.unwrap().unwrap();
            let reply_body = if let MsgBodyOwned::Read { offset, size } = msg.body() {
                let offset: usize = (*offset).try_into().unwrap();
                let size: usize = (*size).try_into().unwrap();
                let end: usize = offset + size;
                MsgBodyRef::Write {
                    offset: offset as u64,
                    buf: &data[offset..end],
                }
            } else {
                MsgBodyRef::Fail(&MsgError::Unknown)
            };
            sender_two
                .send_msg(&block_path_one, reply_body)
                .await
                .unwrap();
        });

        sender_one
            .send_msg(&block_path_two, MsgBodyRef::Read { offset: 2, size: 2 })
            .await
            .unwrap();
        handle.await.unwrap();
        let reply = receiver_two.next().await.unwrap().unwrap();
        if let MsgBodyOwned::Write { offset, buf } = reply.body() {
            assert_eq!(2, *offset);
            assert_eq!([2, 3].as_slice(), buf.as_slice());
        } else {
            panic!("replay was not the right type");
        };
    }
}
